/*
 * osc.c
 * Kipton Moravec 6/30/2017
 *
 *
 * This sets up the clock section of the processor
 *
 * XT2CLK : a 14745600Hz crystal
 * XT1CLK : 32768 Hz
 * VLFO   : 6 - 9.4 - 14 kHz
 * REFO   : 32768 Hz
 *
 * ACLK   : 32768 Hz used by clocks
 *
 *
 * MCLK   : 14.74560 MHz
 * used by CPU
 *
 *
 * SMCLK   : 14745600Hz
 * used by UART, and SPI and maybe I2C
 * Later we may try a division of the 14 MHz by 2 or 4 to see if we save power
 *
 * So for now we are not going to use the FLL.
 * And we will start up XT1 and use it on ACLK for now
 * we may switch to REFO later.
 *
 */
//#define DEBUG_OSC
#define ACLK_XT1CLK
//#define USE_FLL


#include <msp430.h>
#include "osc.h"

//#include <msp430F5329.h>
//#include <io430f5329.h>
#include <stdint.h>

void CLK_Init(void)
{
#ifdef DEBUG_OSC
  P2DIR |= BIT2;                            // SMCLK set out to pins P2.2
  P2SEL |= BIT2;
  P7DIR |= BIT7;                            // MCLK set out to pins P7.7
  P7SEL |= BIT7;
  P1DIR |= BIT0;                            // ACLK set out to pins P1.0
  P1SEL |= BIT0;

#endif

// This sets up the pins
  P5SEL |= XT1OUT | XT1IN | XT2OUT | XT2IN;
                                            // Select XT1   Pin 12 P5.4 XT1IN
                                            //              Pin 13 P5.5 XT1OUT
                                            // Select XT2   Pin 69 P5.2 XT2IN
                                            //              Pin 70 P5.3 XT2OUT

  UCSCTL6 = XT2DRIVE0 | XCAP_3;
                                            // XT2DRIVE = 01 -> 8 MHz - 16 MHz
                                            // XT2BYPASS = 0 External Crystal
                                            // XT2OFF = 0 -> On
                                            // XT1DRIVE = 00 Lowest Current for 4 - 8 MHz
                                            // XTS = 0 Low Frequency mode XT1,bits define capacitance
                                            // XT1BYPASS= 0 external crystal
                                            // XCAP = 3 5 pF / 2 = 2.5 pF
                                            // SMCLKOFF = 0 On
                                            // XT1OFF = 0 On

  UCSCTL5 = DIVPA_DIV1 | DIVA_DIV1 | DIVS_DIV1 | DIVM_DIV1;
                                            // DIVPA = 0 /1
                                            // DIVA =  0 /1
                                            // DIVS =  0 /1
                                            // DIVM =  0 /1


  UCSCTL3 = 0;                              // FLL Reference Clock = XT1
                                            // and divide input by 1

  // Loop until XT1,XT2 & DCO stabilizes - In this case loop until XT1 and DCo settle
  do
  {
    UCSCTL7 &= ~(XT2OFFG + XT1LFOFFG + DCOFFG);
                                            // Clear XT2,XT1,DCO fault flags
    SFRIFG1 &= ~OFIFG;                      // Clear fault flags

  } while (SFRIFG1&OFIFG);                  // Test oscillator fault flag




  //UCSCTL6_bit.XT1DRIVE1 = 0;
  UCSCTL6 &= ~(XT1DRIVE_3);                 // Xtal is now stable, reduce drive strength

#ifdef ACLK_XT1CLK
  UCSCTL4 = 0x0055;                         // ACLK = XT1CLK  // external Crystal
                                            // SMCLK = XT2CLK
                                            // MCLK = XT2CLK
#else
  UCSCTL4 = 0x0255;                         // ACLK = REFOCLK // internal Oscillator
                                            // SMCLK = XT2CLK
                                            // MCLK = XT2CLK
#endif


  UCSCTL3 = SELREF_2;                       // Set DCO FLL reference = REFO
//  UCSCTL4 |= SELA_2;                        // Set ACLK = REFO


  __bis_SR_register(SCG0);                  // Disable the FLL control loop

#ifdef USE_FLL
  UCSCTL0 = 0x0000;                         // Set lowest possible DCOx, MODx
  UCSCTL1 = DCORSEL_5;                      // Select DCO range 16MHz operation
  UCSCTL2 = FLLD_1 + 249;                   // Set DCO Multiplier for 8MHz
                                            // (N + 1) * FLLRef = Fdco
                                            // (249 + 1) * 32768 = 8MHz

  __bic_SR_register(SCG0);                  // Enable the FLL control loop


  // Worst-case settling time for the DCO when the DCO range bits have been
  // changed is n x 32 x 32 x f_MCLK / f_FLL_reference. See UCS chapter in 5xx
  // UG for optimization.
  // 32 x 32 x 8 MHz / 32,768 Hz = 250000 = MCLK cycles for DCO to settle
  __delay_cycles(250000);
#endif

  // Loop until XT1,XT2 & DCO stabilizes - In this case only DCO has to stabilize
  do
  {
    UCSCTL7 &= ~(XT2OFFG + XT1LFOFFG + DCOFFG);
                                            // Clear XT2,XT1,DCO fault flags
    SFRIFG1 &= ~OFIFG;                      // Clear fault flags
  }while (SFRIFG1&OFIFG);                   // Test oscillator fault flag

}

void speed_10K(void)
{
  uint16_t t;
  t = UCSCTL4;
  t &= SELM_CLR;        // clears SELM field to 000
  t |= SELM_VLOCLK;     // sets to VLOCLK source
  UCSCTL4 = t;          // make change
}

void speed_32K(void)
{
  uint16_t t;
  t = UCSCTL4;
  t &= SELM_CLR;        // clears SELM field to 000
  t |= SELM_REFOCLK;    // sets to REFOCLK source
  UCSCTL4 = t;          // make change
}

void full_speed(void)
{
  uint16_t t;
  UCSCTL5 &= DIVM_CLR;  // sets to DIV1 or 14.74560 MHz

  t = UCSCTL4;
  t &= SELM_CLR;        // clears SELM field to 000
  t |= SELM_XT2CLK;     // sets to X2TCLK source
  UCSCTL4 = t;          // make change
}

void reduced_speed(void)
{
  uint16_t t;
  t = UCSCTL5;
  t &= DIVM_CLR;        // clears DIVM field to 000
  t |= DIVM_DIV16;      // sets to DIV16 or 921600 Hz
  UCSCTL5 = t;          // make change

  t = UCSCTL4;
  t &= SELM_CLR;        // clears SELM field to 000
  t |= SELM_XT2CLK;     // sets to X2TCLK source
  UCSCTL4 = t;          // make change
}

