/*
 * usi_i2c.c
 * Kipton Moravec 6/30/2017
 *
 *
 * This has been modified bacause the names have been changed from the version
 * Jan was using to the MSP430F5329.
 *
 * I like that he used the interrupt to operate this in the background, and
 * this is why I started with his code versus someone elses. Plus his code was
 * small
 *
 * The beginning I will test the bq34z100-g1 with a write address of 0xAA and a
 * read address of 0xAB.
 *
 * XT2CLK : a 14745600Hz crystal
 * XT1CLK : a 8 MHZ clock
 *        : 32768 Hz
 * VLFO   : 6 - 9.4 - 14 kHz
 * REFO   : 32768 Hz
 *
 */
/*
    msp430_usi_i2c

    This is a small library that implements an I2C (IIC, IC, or I squared C)
    master on MSP430 devices that only have the USI module (for example,
    MSP430G2412 or MSP430G2452).

    License

    MIT. I believe in freedom, which means I believe in letting you do
    whatever you want with this code.

    Features:
    Small.
    Works.
    Reads and writes.
    Implements repeated start.
    Uses the Bus Pirate convention.

    Rationale:
    I wrote this out of frustration. There is lots of code floating around,
    most of which I didn't like. TI supplies examples which seem to have been
    written by an intern and never looked at again. The examples are overly
    complex, unusable in practical applications, ugly and badly formatted, and
    sometimes even incorrect.

    The MSP430G2xx2 devices are tiny and inexpensive and could be used in many
    application requiring I2C, but many people avoid them because it is so
    annoyingly difficult to use I2C with the USI module.

    This code is very, very loosely based on the msp430g2xx2_usi_16.c example
    from TI, but if you compare you will notice that:

    the state machine is different (simpler): see doc/usi-i2c-state-diagram.pdf
    for details,
    it actually has a useful interface,
    it is smaller.

    Limitations:
    This is a simple I2C master that needs to fit on devices that have 128
    bytes of RAM, so scale your expectations accordingly. There is no error
    detection, no arbitration loss detection, only master mode is implemented.
    Addressing is fully manual: it is your responsibility to shift the 7-bit
    I2C address to the left and add the R/W bit (actually, I see this as an
    advantage).

    Usage

    There are two functions: i2c_init() and i2c_send_sequence().

    You call i2c_init() once to initialize the USI module. You have to provide
    the constants used to configure the USI clock: one of the USIDIV_*
    constants as a usi_clock_divider parameter, which will set the clock divider
    used for USI I2C communications. The usi_clock_source parameter should be
    set to one of the USISSEL* constants. As an example,
    i2c_init(USIDIV_5, USISSEL_2) uses SMCLK/16.

    Data transmission (both transmit and receive) is handled by
    i2c_send_sequence(). It sends a command/data sequence that can include
    restarts, writes and reads. Every transmission begins with a START, and
    ends with a STOP so you do not have to specify that.

    i2c_send_sequence() will busy-spin if another transmission is in progress.
    Note that this is interrupt-driven asynchronous code: you can't just call
    i2c_send_sequence from an interrupt handler and expect it to work: you risk
    a deadlock if another transaction is in progress and nothing will happen
    before the interrupt handler is done anyway. So the proper way to use this
    is in normal code. This should not be a problem, as performing such lengthy
    tasks as I2C communication inside an interrupt handler is a bad idea anyway.
    wakeup_sr_bits should be a bit mask of bits to clear in the SR register when
    the transmission is completed (to exit LPM0: LPM0_BITS (CPUOFF), for LPM3:
    LPM3_bits (SCG1+SCG0+CPUOFF))

    i2c_send_sequence() takes four parameters:

    handle is the handle returned from i2c_open()

    sequence is the I2C operation sequence that should be performed. It can
    include any number of writes, restarts and reads. Note that the sequence is
    composed of uint16_t, not uint8_t elements. This is because we have to
    support out-of-band signalling of I2C_RESTART and I2C_READ operations,
    while still passing through 8-bit data.

    sequence_length is the number of sequence elements (not bytes). Sequences
    of length up to 65535 are supported.

    received_data should point to a buffer that can hold as many bytes as there
    are I2C_READ operations in the sequence. If there are no reads, 0 can be
    passed, as this parameter will not be used.

    i2c_send_sequence() uses the Bus Pirate I2C convention, which I found to
    be very useful and compact. As an example, this Bus Pirate sequence:

    "[0x38 0x0c [ 0x39 r ]"
    is specified as:

    {0x38, 0x0c, I2C_RESTART, 0x39, I2C_READ};
    in I2C terms, this sequence means:

    Write 0x0c to device 0x1c (0x0c is usually the register address).
    Do not release the bus.
    Issue a repeated start.
    Read one byte from device 0x1c (which would normally be the contents of
    register 0x0c on that device).
    The sequence may read multiple bytes:

    {0x38, 0x16, I2C_RESTART, 0x39, I2C_READ, I2C_READ, I2C_READ};
    This will normally read three bytes from device 0x1c starting at register
    0x16. In this case you need to provide a pointer to a buffer than can hold
    three bytes.

    Note that start and stop are added for you automatically, but addressing
    is fully manual: it is your responsibility to shift the 7-bit I2C address
    to the left and add the R/W bit. The examples above communicate with a
    device whose I2C address is 0x1c, which shifted left gives 0x38. For reads
    we use 0x39, which is (0x1c<<1)|1.

    If you wonder why I consider the Bus Pirate convention useful, note that
    what you specify in the sequence is very close to the actual bytes on the
    wire. This makes debugging and reproducing other sequences easy. Also, you
    can use the Bus Pirate to prototype, and then easily convert the tested
    sequences into actual code.

    Steps to use this code:

    Add the files to your project.

    Call i2c_init() with appropriate USI settings:

    i2c_init(USIDIV_5, USISSEL_2);

    Communicate. Here's an example of performing a read with repeated start
    restart) from an MMA845x accelerometer. Note that the chip goes into LPM0
    sleep while I2C transmission is interrupt-driven. We will get woken up
    after the transmit/receive is done, but we still check i2c_done() just in
    case something else woke us up.

    uint16_t mma8453_read_interrupt_source[] = {0x38, 0x0c, I2C_RESTART, 0x39, I2C_READ};
    uint8_t status;
    i2c_send_sequence(mma8453_read_interrupt_source, 5, &status, LPM0_BITS);
    LPM0;
    while(!i2c_done());
    Does it work?

    It does for me: I've been using this code in a number of projects and had
    no problems with it. I've used it on MSP430G2412 and MSP430G2452 chips.
    That said, there are no guarantees.
*/

/*
  usi_i2c.c

  Copyright (C) 2013 Jan Rychter

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#define testbq796
#include <msp430.h>
#include <stdint.h>
#include <intrinsics.h>

#include "usi_i2c.h"

void start_communications(void);

// Internal state
static uint16_t const *i2c_sequence;
static uint16_t i2c_sequence_length;
static uint8_t *i2c_receive_buffer;
static uint16_t i2c_wakeup_sr_bits;

i2c_state_type i2c_state = I2C_IDLE;

volatile uint8_t result_UCB1IFG;
volatile uint16_t i2c_interrupt_count;

static inline void i2c_prepare_stop();
static inline void i2c_prepare_data_xmit_recv();

#ifdef testbq796
/*
 * Theis tests i2C
 */
#define bq769x_Address_WR_t  0x30
#define bq769x_Address_RD_t  0x31
const uint16_t i2c_testdata_sequence1[] = {  bq769x_Address_WR_t, 0, I2C_RESTART, bq769x_Address_RD_t, I2C_READ, I2C_READ, I2C_READ };

#define RDL 10
uint8_t receive_data[RDL];


static void test_i2c(void)
{
  uint16_t sequence_length = 7;
  volatile uint32_t dummy;

  uint16_t i;
  //uint16_t j;

  for (i=0; i<RDL; i++)
  {
    receive_data[i] = 0xAA;
  }

  i2c_send_sequence(i2c_testdata_sequence1,sequence_length,receive_data,0);

  start_communications();

}

#endif


#define DATA_SIZE 10
uint16_t dw, dr; // index pointers
uint8_t data_written[DATA_SIZE];
uint8_t data_read[DATA_SIZE];

void clear_data(void)
{
  uint16_t i;
  for (i=0; i<DATA_SIZE;i++)
  {
    data_written[i] = 0xAA;
    data_read[i] = 0xAA;
  }
  dw = 0;
  dr = 0;
}




void i2c_send_sequence(uint16_t const * sequence, uint16_t sequence_length, uint8_t *received_data, uint16_t wakeup_sr_bits)
{
  while(i2c_state != I2C_IDLE); // we can't start another sequence until the current one is done
  i2c_sequence = sequence;
  i2c_sequence_length = sequence_length;
  i2c_receive_buffer = received_data;
  i2c_wakeup_sr_bits = wakeup_sr_bits;
  i2c_state = I2C_START;
  //UCB1IFG |= UCTXIFG;            // actually start communication
}
volatile uint8_t indata;

static inline void i2c_prepare_stop()
{
  //USICTL0 |= USIOE;             // SDA = output
  //UCB1TXBUF = 0x00;
  //USICNT |=  0x01;              // Bit counter= 1, SCL high, SDA low

  UCB1CTL1 |= UCTR | UCTXSTP;     // Transmitter and Stop
  i2c_state = I2C_STOP;
}

static inline void i2c_prepare_data_xmit_recv(void)
{
  indata = *i2c_sequence;

  if(i2c_sequence_length == 0)
  {
    i2c_prepare_stop();         // nothing more to do, prepare to send STOP
  }
  else
  {
    if(*i2c_sequence == I2C_RESTART)
    {
      UCB1CTL1 |= UCTR;
      //USICTL0 |= USIOE;       // SDA = output
      //UCB1TXBUF = 0xff;       // prepare and send a dummy bit, so that SDA is high
      //USICNT = (USICNT & 0xE0) | 1;
      i2c_state = I2C_START;
    }
//    else if(*i2c_sequence == I2C_RESTART)

    else if(*i2c_sequence == I2C_READ)
    {
      UCB1CTL1 &= ~(UCTR);   // Receive
      //USICTL0 &= ~USIOE;               // SDA = input
      //USICNT = (USICNT & 0xE0) | 8;    // Bit counter = 8, RX data
      i2c_state = I2C_RECEIVED_DATA;   // next state: Test data and ACK/NACK
    }
    else
    {                           // a write
      // at this point we should have a pure data byte, not a command, so (*i2c_sequence >> 8) == 0
      UCB1CTL1 |= UCTR;                 // Transmit Mode
      //USICTL0 |= USIOE;               // SDA = output

      UCB1TXBUF = (char)(*i2c_sequence);// Load data byte
      //USICNT = (USICNT & 0xE0) | 8;   // Bit counter = 8, start TX
      i2c_state = I2C_PREPARE_ACKNACK;  // next state: prepare to receive data ACK/NACK
    }
    i2c_sequence++;
    i2c_sequence_length--;
  }
}

void i2c_init(void)
{
#ifdef TEST_I2C_SIGNALS
  P4SEL &= 0xF9;
  P4DIR |= 0x06;
 `while(1)
  {
    P4OUT |= 0x02;
    P4OUT |= 0x06;
    P4OUT &= 0xFD;
    P4OUT &= 0xF9;
  }
#endif

  // Setup pins
  P4SEL |= 0x06;            // Setup as P4.1 (SDA), P4.2(SCL) as Special Function pins

  __disable_interrupt();
  i2c_interrupt_count = 0;
  UCB1CTL1 = 0x81;          // Clock is SMCLK, Do not interrupt on erroneous character
                            // do not interrupt on break character
                            // Not dormant, next frame is data, next frame is not a break,
                            // I2C is held in reset.

  UCB1BR0 = 148;            // Set bit rate to 14745600 / 148 = 99632.43 bps
  UCB1BR1 = 0;

  UCB1CTL0 = 0x0F;          // 7 Bit addressing, Single Master
                            // Master mode select, I2C mode, Synchronous Mode

  UCB1CTL1 = 0x81;          // Clock is SMCLK, Do not interrupt on erroneous character
                            // do not interrupt on break character
                            // Not dormant, next frame is data, next frame is not a break,
                            // I2C is not held in reset.
  i2c_state = I2C_IDLE;         // Make sure state machine is in idle
  UCB1CTL1 &= ~(UCSWRST);       // Clear I2C Reset now I2C is setup and ready.


  UCB1IE = 0x03;            // Enable Transmit and Receive Interrupt
  UCB1IFG = 0;              // Clear all interrupt flags.



  //UCB1CTL0 = UCMST|USISWRST;  // Master Port & USI mode setup
  //USICTL1 = USII2C|USIIE;                   // Enable I2C mode & USI interrupt
  //USICKCTL = usi_clock_divider | usi_clock_source | USICKPL;
  //USICNT |= USIIFGCC;                       // Disable automatic clear control
  //USICTL0 &= ~USISWRST;                     // Enable USI
  //USICTL1 &= ~USIIFG;                       // Clear pending flag



  //__enable_interrupt();

#ifdef testbq796
  test_i2c();
#endif
}

#if 0
#if 1
#pragma vector = USCI_B1_VECTOR
__interrupt void USCI_I2C_unterrupt(void)
{
  i2c_interrupt_count++;
  switch(__even_in_range(i2c_state,12)) // defined in EW430_Compiler_Reference p354
  {                                     // makes switch statement into computed goto
  case I2C_IDLE: // 0

    break;

  case I2C_START:                   // generate start condition -- 2

    UCB1CTL1 |= UCTR | UCTXSTT;     // Transmitter and Transmit START
    UCB1I2CSA = ((char)(*i2c_sequence)) >> 1; // Load I2C Address without r/w bit byte
    i2c_sequence++;
    i2c_sequence_length--;
    //UCB1TXBUF = 0x00;             // I think it is discarded
    //USICTL0 |= (USIGE|USIOE);
    //USICTL0 &= ~USIGE;
    i2c_prepare_data_xmit_recv();
    i2c_state = I2C_PREPARE_ACKNACK; // next state: prepare to receive data ACK/NACK
    break;

  case I2C_PREPARE_ACKNACK:         // prepare to receive ACK/NACK -- 4
    UCB1CTL1 &= ~(UCTR);            // Receive
    //USICTL0 &= ~USIOE;            // SDA = input
    //USICNT |= 0x01;               // Bit counter=1, receive (N)Ack bit
    i2c_state = I2C_HANDLE_RXTX;    // Go to next state: check ACK/NACK and continue xmitting/receiving if necessary
    break;

  case I2C_HANDLE_RXTX:         // Process Address Ack/Nack & handle data TX -- 6
    if((UCB1IFG & UCSTPIFG) != 0)
    {  // did we get a NACK?
      i2c_prepare_stop();
    } else {
      i2c_prepare_data_xmit_recv();
    }
    break;

  case I2C_RECEIVED_DATA:       // received data, send ACK/NACK -- 8
    *i2c_receive_buffer = UCB1RXBUF;
    i2c_receive_buffer++;
    UCB1CTL1 |= UCTR;           // Transmit
    //USICTL0 |= USIOE;           // SDA = output
    if(i2c_sequence_length > 0)
    {
      // If this is not the last byte
      UCB1TXBUF = 0x00;             // ACK
      i2c_state = I2C_HANDLE_RXTX;  // Go to next state: data/rcv again
    } else {                        // last byte: send NACK
      UCB1TXBUF = 0xff;                // NACK
      i2c_state = I2C_PREPARE_STOP; // stop condition is next
    }
    //USICNT |= 0x01;             // Bit counter = 1, send ACK/NACK bit
    break;

  case I2C_PREPARE_STOP:        // prepare stop condition -- 10
    i2c_prepare_stop();         // prepare stop, go to state 14 next
    break;

  case I2C_STOP:                // Generate Stop Condition -- 12
    UCB1TXBUF = 0x0FF;          // UCB1TXBUF = 1 to release SDA
    //UCB1CTL1 |= UCTR | UCTXSP;
    //USICTL0 |= USIGE;         // Transparent latch enabled
    //USICTL0 &= ~(USIGE|USIOE);  // Latch/SDA output disabled
    i2c_state = I2C_IDLE;       // Reset state machine for next xmt
    if(i2c_wakeup_sr_bits)
    {
        __bic_SR_register_on_exit(i2c_wakeup_sr_bits); // exit active if prompted to
    }
    break;
  }
  //USICTL1 &= ~USIIFG;         // Clear pending flag
  UCB1IFG &= 0xFFFC;            // Clear UCTXIFG and UCRXIFG interrupt flags
}
#else

#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector = USCI_B1_VECTOR
__interrupt void USCI_I2C_interrupt(void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(USCI_B0_VECTOR))) USCI_B0_ISR (void)
#else
#error Compiler not supported!
#endif
{
  switch(__even_in_range(UCB0IV,12))
  {
  case  0: break;                           // Vector  0: No interrupts
  case  2: break;                           // Vector  2: ALIFG
  case  4: break;                           // Vector  4: NACKIFG
  case  6: break;                           // Vector  6: STTIFG
  case  8: break;                           // Vector  8: STPIFG
  case 10: break;                           // Vector 10: RXIFG
  case 12:                                  // Vector 12: TXIFG
    if (TXByteCtr)                          // Check TX byte counter
    {
      UCB1TXBUF = *PTxData++;               // Load TX buffer
      TXByteCtr--;                          // Decrement TX byte counter
    }
    else
    {
      UCB1CTL1 |= UCTXSTP;                  // I2C stop condition
      UCB1IFG &= ~UCTXIFG;                  // Clear USCI_B0 TX int flag
     // __bic_SR_register_on_exit(LPM0_bits); // Exit LPM0
    }
  default: break;
  }
}
#endif

#endif

void start_communications(void)
{
    uint16_t i;
    clear_data();
    UCB1CTL1 |= UCTR | UCTXSTT;     // Transmitter and Transmit START
    data_written[dw] = ((char)(*i2c_sequence)) >> 1;
    UCB1I2CSA = ((char)(*i2c_sequence)) >> 1; // Load I2C Address without r/w bit byte

    i2c_sequence++;
    i2c_sequence_length--;

//    i2c_prepare_data_xmit_recv();
    i2c_state = I2C_PREPARE_ACKNACK; // next state: prepare to receive data ACK/NACK

    while(UCB1IFG == 0);

    for (i=0;i<10; i++);

}
